// z80codetodisc.cpp
// Revision 5-oct-2003

#include "z80snap.h"

#include <string>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <map>

const std::string progname ("z80codetodisc");
const std::string version ("1.0.0\n(C) 2003 Julian Albo");

class BadArgs { };

int doit (int argc, char * argv [] )
{
	using std::cout;
	using std::cerr;
	using std::endl;
	using z80snap::CHAR;

	// Analyze options.

	if (argc >= 2 && (strcmp (argv [1], "-version") == 0) )
	{
		cout << progname << " version " << version << endl;
		return 0;
	}

	if (argc != 5 && argc != 4)
		throw BadArgs ();

	char * filein= argv [1];
	char * fileout= argv [2];
	size_t position;
	size_t length;
	if (strcmp (argv [3], "SCREEN$") == 0)
	{
		if (argc != 4)
			throw BadArgs ();
		position= 16384;
		length= 6912;
	}
	else
	{
		if (argc != 5)
			throw BadArgs ();
		position= atoi (argv [3] );
		length= atoi (argv [4] );
	}

	// Load snapshot.

	std::ifstream is (filein, std::ios::in | std::ios::binary);
	if (! is.is_open () )
		throw "Can't open input file";

	z80snap::Snapshot snap (is);
	is.close ();

	// PLUS3DOS header.

	CHAR plus3 [128] = {
		'P', 'L', 'U', 'S', '3', 'D', 'O', 'S', // Identifier.
		0x1A, // CP/M soft end of file.
		1,    // Issue number.
		0,    // Version number.
	};
	size_t filesize= length + 128;
	plus3 [11]= filesize & 0xFF;
	plus3 [12]= (filesize >> 8) & 0xFF;
	plus3 [13]= (filesize >> 16) & 0xFF;
	plus3 [14]= (filesize >> 24) & 0xFF;
	plus3 [15]= 3; // type: CODE.
	// Length.
	plus3 [16]= length & 0xFF;
	plus3 [17]= (length >> 8) & 0xFF;
	// Position.
	plus3 [18]= position & 0xFF;
	plus3 [19]= (position >> 8) & 0xFF;
	// Don't know if used for something.
	plus3 [20]= 0x80;
	plus3 [21]= 0x80;
	// Compute checksum
	CHAR check= 0;
	for (int i= 0; i < 127; ++i)
		check+= plus3 [i];
	plus3 [127]= check;

	// Write the file.

	std::ofstream os (fileout, std::ios::out | std::ios::binary);
	if (! os.is_open () )
		throw "Can't open output file";
	os.write (reinterpret_cast <char *> (plus3), sizeof (plus3) );
	if (! os)
		throw "Error writing";
	os.write (reinterpret_cast <char *> (snap.basememory () + position),
		length);
	size_t round= 128 - (length % 128);
	if (round != 128)
	{
		char aux [128]= {};
		os.write (aux, round);
	}

	return 0;
}

int main (int argc, char * argv [] )
{
	using std::cout;
	using std::cerr;
	using std::endl;
	using std::flush;

	try
	{
		return doit (argc, argv);
	}
	catch (BadArgs &)
	{
		cerr << "Usage: " << progname << " "
			"file.z80 file_code position length\n\t"
			"or: " << progname << " "
			"file.z80 file_code SCREEN$\n\t" <<
			"or: " << progname << " -version" <<
			endl;
		return 1;
	}
	catch (std::exception & e)
	{
		cout << flush;
		cerr << "ERROR: " << e.what () << endl;
		return 1;
	}
	catch (const char * str)
	{
		cout << flush;
		cerr << "ERROR: " << str << endl;
		return 1;
	}
}


// End of z80codetodisc.cpp
