// z80list.cpp
// Revision 5-oct-2003

#include "z80snap.h"

#include <string>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <sstream>
#include <stdlib.h>
#include <map>

const std::string progname ("z80list");
const std::string version ("1.0.0\n(C) 2003 Julian Albo");

//	Spectrum Basic keywords.

typedef std::map <z80snap::CHAR, std::string> keyword_t;
keyword_t keyword;

void initkeyword ()
{
	keyword [165]= "RND";
	keyword [166]= "INKEY$";
	keyword [167]= "PI";
	keyword [168]= "FN ";
	keyword [169]= "POINT ";
	keyword [170]= "SCREEN$ ";
	keyword [171]= "ATTR ";
	keyword [172]= "AT ";
	keyword [173]= "TAB ";
	keyword [174]= "VAL$ ";
	keyword [175]= "CODE ";
	keyword [176]= "VAL ";
	keyword [177]= "LEN ";
	keyword [178]= "SIN ";
	keyword [179]= "COS ";
	keyword [180]= "TAN ";
	keyword [181]= "ASN ";
	keyword [182]= "ACS ";
	keyword [183]= "ATN ";
	keyword [184]= "LN ";
	keyword [185]= "EXP ";
	keyword [186]= "INT ";
	keyword [187]= "SQR ";
	keyword [188]= "SGN ";
	keyword [189]= "ABS ";
	keyword [190]= "PEEK ";
	keyword [191]= "IN ";
	keyword [192]= "USR ";
	keyword [193]= "STR$ ";
	keyword [194]= "CHR$ ";
	keyword [195]= "NOT ";
	keyword [196]= "BIN ";
	keyword [197]= " OR ";
	keyword [198]= " AND ";
	keyword [199]= "<=";
	keyword [200]= ">=";
	keyword [201]= "<>";
	keyword [202]= " LINE ";
	keyword [203]= " THEN ";
	keyword [204]= " TO ";
	keyword [205]= " STEP ";
	keyword [206]= " DEF FN ";
	keyword [207]= " CAT ";
	keyword [208]= " FORMAT ";
	keyword [209]= " MOVE ";
	keyword [210]= " ERASE ";
	keyword [211]= " OPEN #";
	keyword [212]= " CLOSE #";
	keyword [213]= " MERGE ";
	keyword [214]= " VERIFY ";
	keyword [215]= " BEEP ";
	keyword [216]= " CIRCLE ";
	keyword [217]= " INK ";
	keyword [218]= " PAPER ";
	keyword [219]= " FLASH ";
	keyword [220]= " BRIGHT ";
	keyword [221]= " INVERSE ";
	keyword [222]= " OVER ";
	keyword [223]= " OUT ";
	keyword [224]= " LPRINT ";
	keyword [225]= " LLIST ";
	keyword [226]= " STOP ";
	keyword [227]= " READ ";
	keyword [228]= " DATA ";
	keyword [229]= " RESTORE ";
	keyword [230]= " NEW ";
	keyword [231]= " BORDER ";
	keyword [232]= " CONTINUE ";
	keyword [233]= " DIM ";
	keyword [234]= " REM ";
	keyword [235]= " FOR ";
	keyword [236]= " GO TO ";
	keyword [237]= " GO SUB ";
	keyword [238]= " INPUT ";
	keyword [239]= " LOAD ";
	keyword [240]= " LIST ";
	keyword [241]= " LET ";
	keyword [242]= " PAUSE ";
	keyword [243]= " NEXT ";
	keyword [244]= " POKE ";
	keyword [245]= " PRINT ";
	keyword [246]= " PLOT ";
	keyword [247]= " RUN ";
	keyword [248]= " SAVE ";
	keyword [249]= " RANDOMIZE ";
	keyword [250]= " IF ";
	keyword [251]= " CLS ";
	keyword [252]= " DRAW ";
	keyword [253]= " CLEAR ";
	keyword [254]= " RETURN ";
	keyword [255]= " COPY ";
}

void additionalkeywords ()
{
	keyword [163]= " SPECTRUM ";
	keyword [164]= " PLAY ";
}

void nospectrumkeywords ()
{
	// GOTO and GOSUB without space.
	keyword [236]= " GOTO ";
	keyword [237]= " GOSUB ";
	// Trigonometric arc functions.
	keyword [181]= "ASIN ";
	keyword [182]= "ACOS ";
	keyword [183]= "ATAN ";
	// Logarithm.
	keyword [184]= "LOG ";
	// ASC instead of CODE.
	keyword [175]= "ASC ";
	// CONTINUE abreviated.
	keyword [232]= " CONT ";
	// BIN substitued by prefix.
	keyword [196]= "&x";
}

void blassickeywords ()
{
	// DRAWARC instead of DRAW.
	keyword [252]= " DRAWARC ";
}

// Controls that can be embedded in lines with
// size of his parameters.

typedef std::map <z80snap::CHAR, size_t> embed_t;
embed_t embed;

void initembed ()
{
	embed [14]= 5; // Binary value of numeric constant
	embed [16]= 1; // INK
	embed [17]= 1; // PAPER
	embed [18]= 1; // FLASH
	embed [19]= 1; // BRIGHT
	embed [20]= 1; // INVERSE
	embed [21]= 1; // OVER
	embed [22]= 2; // AT
	embed [23]= 2; // TAB
}

class BadArgs { };

std::string encodechar (z80snap::CHAR c)
{
	if (c == '"')
		return "\"\"";
	else if (c < 32)
		return std::string (1, '^') + char (c + 'A' - 1);
	else if (c == '^')
		return "\\^";
	else if (c == '\\')
		return "\\\\";
	else if (c >= 0x80 && c <= 0x9F)
	{
		std::ostringstream oss;
		oss << '[' << std::hex << int (c) << ']';
		return oss.str ();
	}
	else if (c == '[')
		return "\\[";
	else return std::string (1, c);
}

void dumpvariables (z80snap::Snapshot & snap, size_t initvar)
{
	using std::cout;
	using std::endl;

	cout << "Variables area dump." << endl;
	z80snap::CHAR v;
	while ( (v= snap.peek (initvar) ) != 0x80)
	{
		z80snap::CHAR c= (v & 0x1F) + 0x60;
		// First letter of variable name.

		switch (v & 0xE0) // Three more significant bits.
		{
		case 0x60: // 011 Numeric one letter name.
			cout << c << "= " <<
				snap.peeknumber (++initvar) <<
				endl;
			initvar+= 5;
			break;
		case 0xA0: // 101 Numeric name longer than one.
			cout << c;
			while ( (c= snap.peek (++initvar) ) < 0x80)
				cout << c;
			cout << z80snap::CHAR (c & 0x7F) << "= " <<
				snap.peeknumber (++initvar) <<
				endl;
			initvar+= 5;
			break;
		case 0x80: // 100 Numeric array.
			{
				size_t l= snap.peek16 (++initvar);
				unsigned int ndims= snap.peek (initvar + 2);
				cout << c << '(';
				unsigned int nelems= 1;
				for (unsigned int i= 0; i < ndims; ++i)
				{
					unsigned int dim= snap.peek16
						(initvar + 3 + 2 * i);
					cout << dim;
					nelems*= dim;
					if (i < ndims - 1)
						cout << ',';
				}
				cout << ")= ";
				unsigned int base= initvar + 3 + 2 * ndims;
				for (unsigned int i= 0; i < nelems; ++i)
				{
					cout << snap.peeknumber (base + 5 * i);
					if (i < nelems - 1)
						cout << ',';
				}
				cout << endl;
				initvar+= l + 2;
			}
			break;
		case 0xE0: // 111 For/next control variable.
			cout << "FOR " << c << "= " <<
				snap.peeknumber (++initvar) << " TO ";
			// Need a sequence point.
			cout << snap.peeknumber (initvar + 5) << " STEP " <<
				snap.peeknumber (initvar + 10) << " LINE " <<
				snap.peek16 (initvar + 15) << ':' <<
				int (snap.peek (initvar + 17) ) << endl;
			initvar+= 18;
			break;
		case 0x40: // 010 String.
			{
				size_t l= snap.peek16 (++initvar);
				cout << c << "$= \"";
				for (size_t i= 0; i < l; ++i)
				{
					c= snap.peek (initvar + 2 + i);
					cout << encodechar (c);
				}
				cout << '"' << endl;
				initvar+= 2 + l;
			}
			break;
		case 0xC0: // 110 Character array.
			{
				size_t l= snap.peek16 (++initvar);
				unsigned int ndims= snap.peek (initvar + 2);
				cout << c << "$(";
				unsigned int nelems= 1;
				for (unsigned int i= 0; i < ndims; ++i)
				{
					unsigned int dim= snap.peek16
						(initvar + 3 + 2 * i);
					cout << dim;
					nelems*= dim;
					if (i < ndims - 1)
						cout << ',';
				}
				cout << ")= \"";
				unsigned int base= initvar + 3 + 2 * ndims;
				for (unsigned int i= 0; i < nelems; ++i)
				{
					c= snap.peek (base + i);
					cout << encodechar (c);
				}
				cout << '"' << endl;
				initvar+= l + 2;
			}
			break;
		default:
			throw "Variable area corrupt";
		}
	}
	cout << "End of variables dump." << endl;
}

int doit (int argc, char * argv [] )
{
	using std::cout;
	using std::cerr;
	using std::endl;
	using z80snap::CHAR;

	if (argc <= 1)
		throw BadArgs ();
	bool shownumbers= false;
	bool showinfo= false;
	bool usecr= false;
	bool showchr= false;
	bool showembed= false;
	bool showchrgraph= false;
	bool hidegraph= false;
	bool nospectrum= false;
	bool blassic= false;
	bool debug= false;
	bool dumpvars= false;
	char * filename= NULL;
	size_t endprog= 0;
	size_t addr= 0;

	for (int i= 1; i < argc; ++i)
	{
		char * current= argv [i];
		if (strcmp (current, "-version") == 0)
		{
			cout << progname << " version " << version << endl;
			return 0;
		}
		else if (strcmp (current, "-numbers") == 0)
			shownumbers= true;
		else if (strcmp (current, "-info") == 0)
			showinfo= true;
		else if (strcmp (current, "-cr") == 0)
			usecr= true;
		else if (strcmp (current, "-chr") == 0)
			showchr= true;
		else if (strcmp (current, "-embed") == 0)
			showembed= true;
		else if (strcmp (current, "-chrgraph") == 0)
			showchrgraph= true;
		else if (strcmp (current, "-nograph") == 0)
			hidegraph= true;
		else if (strcmp (current, "-nospectrum") == 0)
			nospectrum= true;
		else if (strcmp (current, "-blassic") == 0)
			blassic= true;
		else if (strcmp (current, "-debug") == 0)
			debug= true;
		else if (strcmp (current, "-dumpvars") == 0)
			dumpvars= true;
		else if (strcmp (current, "-PROG") == 0)
		{
			if (i == argc - 1)
				throw BadArgs ();
			addr= atoi (argv [i + 1] );
			++i;
		}
		else if (strcmp (current, "-VARS") == 0)
		{
			if (i == argc - 1)
				throw BadArgs ();
			endprog= atoi (argv [i + 1] );
			++i;
		}
		else
		{
			if (filename)
				throw BadArgs ();
			filename= argv [i];
		}
		
	}

	initkeyword ();
	if (nospectrum || blassic)
		nospectrumkeywords ();
	if (blassic)
		blassickeywords ();
	initembed ();

	std::ifstream is (filename, std::ios::in | std::ios::binary);
	if (! is.is_open () )
		throw "Can't open file";

	z80snap::Snapshot snap (is, debug);
	const z80snap::Header & head= snap.getheader ();
	if (showinfo)
	{
		z80snap::Regs const & regs= head.getregs ();
		cerr << "PC: " << regs.PC << endl;
		cerr << "SP: " << regs.SP << endl;
		cerr << "Compressed: " <<
			(head.iscompressed () ? "Yes" : "No") << endl;
		cerr << "Machine type: ";
		switch (head.gettype () )
		{
		case z80snap::Type16:
			cerr << "Spectrum 16K";
			break;
		case z80snap::Type48:
			cerr << "Spectrum 48K";
			break;
		case z80snap::Type128:
			cerr << "Spectrum 128K, Plus, Plus 2 or Plus 3";
			break;
		default:
			cerr << "Unknown";
		}
		cerr << endl;
	}

	// Parts of Basic dependant of the type of machine.
	// I don't know in what types of machines the new
	// keywords were introduced. Assume all 128K types.
	if (head.gettype () == z80snap::Type128)
		additionalkeywords ();

	// The Basic program is stored in the area delimited
	// by the system variables PROG and VARS.

	size_t sysvarVARS= 23627;
	size_t sysvarPROG= 23635;

	if (endprog == 0)
		endprog= snap.peek16 (sysvarVARS);
	if (addr == 0)
		addr= snap.peek16 (sysvarPROG);

	if (showinfo)
	{
		cerr << "PROG= " << addr << endl;
		cerr << "VARS= " << endprog << endl;
	}

	while (addr < endprog)
	{
		// Get number and length of the line.
		size_t line= snap.peeklinenumber (addr);
		size_t length= snap.peek16 (addr + 2);
		if (showinfo)
		{
			cerr << "Line number: " << line << endl;
			cerr << "Line length: " << length << endl;
		}

		// LIST the line.
		// Wen using the length of the line instead of
		// the final CR to check the end of line assume
		// the last character is CR and don't list it.
		cout << line;
		size_t i;
		bool instring= false;
		for (i= addr + 4; i < addr + 3 + length; ++i)
		{
			CHAR c= snap.peek (i);
			if (usecr && c == '\r')
				break;
			if (c < 32 && instring && showchr)
			{
				cout << "\"+CHR$ " << int (c) <<
					"+\"";
				continue;
			}
			if (c == '"')
				instring= ! instring;
			keyword_t::iterator it= keyword.find (c);
			if (it != keyword.end () )
				cout << it->second;
			else
			{
				embed_t::iterator it= embed.find (c);
				if (it != embed.end () )
				{
					if (c == 14)
					{
						if (shownumbers)
							cout << '[' << snap.peeknumber (i + 1) << ']';
					}
					else
					{
						if (showembed || instring)
						{
							for (size_t j= i; j <= i + it->second; ++j)
							{
								cout << (j == i ? '[' : ',');
								cout << int (snap.peek (j) );
							}
							cout << ']';
						}
					}
					i+= it->second;
				}
				else
				{
					if (c == 127) // Copyright symbol
						c= 0xA9;
					if (c >= 128 && c <= 164)
					{
						// Graphics and UDG
						if (instring)
						{
							if (showchrgraph)
							{
								cout << "\"+CHR$ " << int (c) << "+\"";
								continue;
							}
						}
						else
						{
							if (hidegraph)
								continue;
						}
					}
					if (c >= 32)
						cout << c;
					else
						if (showembed)
							cout << '^' << c + 'A';
				}
			}
		}
		cout << endl;

		// Jump yo next line
		if (usecr)
			addr= i + 1;
		else
			addr+= length + 4;
	}
	if (dumpvars)
		dumpvariables (snap, endprog);

	return 0;
}

int main (int argc, char * argv [] )
{
	using std::cout;
	using std::cerr;
	using std::endl;
	using std::flush;

	try
	{
		return doit (argc, argv);
	}
	catch (BadArgs &)
	{
		cerr << "Usage: " << progname << " "
			"[ -version ] "
			"[ -numbers ] "
			"[ -info ] "
			"[ -cr ] "
			"[ -chr ] "
			"[ -embed ]\n\t"
			"[ -chrgraph ] "
			"[ -nograph ] "
			"[ -nospectrum ] "
			"[ -blassic ] "
			"[ -dumpvars ]\n\t"
			"[ -PROG position ] "
			"[ -VARS position ]\n\t"
			"file.z80" <<
			endl;
		return 1;
	}
	catch (std::exception & e)
	{
		cout << flush;
		cerr << "\nERROR: " << e.what () << endl;
		return 1;
	}
	catch (const char * str)
	{
		cout << flush;
		cerr << "\nERROR: " << str << endl;
		return 1;
	}
}


// End of z80list.cpp
