// z80snap.h
// Revision 5-oct-2003
// (C) 2003 Julian Albo.

#include <iostream>
#include <cstddef>
#include <stdexcept>

namespace z80snap {

class ErrorReadingHeader : public std::exception {
public:
	const char * what () const throw ()
	{ return "Error reading Z80 header"; }
};

typedef unsigned char CHAR;

enum MachineType { TypeUnknown, Type16, Type48, Type128 };

struct Regs {
	unsigned A;
	unsigned F;
	unsigned BC;
	unsigned DE;
	unsigned HL;
	unsigned Ap;
	unsigned Fp;
	unsigned BCp;
	unsigned DEp;
	unsigned HLp;
	unsigned IX;
	unsigned IY;
	unsigned PC;
	unsigned SP;
};

class Header {
public:
	Header (std::istream & is, bool debug= false);
	bool isoldstyle () const { return oldstyle; }
	bool iscompressed () const { return compressed; }
	MachineType gettype () const { return type; }
	const Regs & getregs () const { return regs; }
private:
	bool oldstyle;
	bool compressed;
	MachineType type;
	Regs regs;
};

class Snapshot {
public:
	Snapshot (std::istream & is, bool debug= false);
	const Header & getheader () const { return head; }
	CHAR peek (size_t n);
	size_t peek16 (size_t n);
	size_t peeklinenumber (size_t n);
	double peeknumber (size_t n);
	const CHAR * basememory () const { return memory; }
	CHAR * basememory () { return memory; }
private:
	Header head;
	void loadblock (std::istream & is,
		CHAR * ptr, std::streamsize length, bool debug);
	void loadblockv1 (std::istream & is, bool debug);
	CHAR memory [64 * 1024 + 1];
	// Extra byte to simplify end marker detection on v1 block.
};

} // namespace z80snap

// End of z80snap.h
